/* xh ver.03 playlist frontend */

#include "xh.h"

/* Prototypes for static functions */
typedef void (*ButtonFunc)(GtkButton *, gpointer);

static GtkWidget *add_button(const gchar *image1,
                 ButtonFunc callback, GtkWidget *container,
                 const gchar *name, gint accel);
typedef void (*MenuFunc)(GtkMenuItem *, gpointer);
static GtkWidget *add_menu_item(const gchar *label, MenuFunc callback,
                gint accel, GtkWidget *sub);

/* Globals */
GtkWidget *window, *frameb, *allbox, *hnbox, *movebox, *eventbox, *statusbar,
    *btbox, *btstop, *btprev, *btplay, *imbox, *implay, *impause, *btnext,
    *btlist, *listbox, *popupmenu, *itemdel, *msep1, *msep2, *msep3, 
    *iteml, *subl, *ckm, *ckl, *itemr, *subr, *filesel = NULL, *filelist,
    *rb0, *rb1, *rb2, *rb3, *imbox2, *imlists, *imlisth;
GtkTreeViewColumn *column;
GtkListStore *filestore;

GtkAccelGroup *accel_group;
int sigchld_pipe[2];
gchar startup_cwd[STRBUFSIZE];

/* Show a message on the status bar */
void status(const char *message) {
    gtk_statusbar_pop(GTK_STATUSBAR(statusbar), contextid);
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), contextid, message);
}

/* Update the statusbar */
void update_window() {
    gchar statbuf[STRBUFSIZE];
    if (paused) {
         snprintf(statbuf, STRBUFSIZE, _("Paused"));
    } else if (playing) {
        snprintf(statbuf, STRBUFSIZE, "%s",
           get_songinfo(last_played)->display_name);
    } else {
        snprintf(statbuf, STRBUFSIZE, _("Stopped."));
    }
    status(statbuf);
}

/* A timeout function to be called after we've had a SIGCHLD. This
   does stuff that we can't do in the signal handler. */
void sigchld_backend(gpointer data, gint source, GdkInputCondition cond) {
    gchar c;
    pid_t deadpid;
    /* Read the character off the pipe */
    read(sigchld_pipe[0], &c, 1);
    /* If the child that has exited isn't the player process, we're not
       interested. */
    deadpid = waitpid(-1, NULL, WNOHANG);
    if (deadpid < 1 || deadpid != childpid)
        return;
    childpid = 0;
    if (playing) {
        if (mode_delete_when_played)
            delete_row(last_played);
        pick_next();
    }
}

/* Called when the child process quits */
void handle_sigchld(int dummy) {
    /* Leave it to GDK to find out about */
    write(sigchld_pipe[1], "!", 1);
}

static GtkTargetEntry target_table[] = {
  { "text/uri-list", 0, 0 },
};

static GtkWidget *add_button(const gchar *image1, ButtonFunc callback,
                  GtkWidget *container, const gchar *name, gint accel) {
    GtkWidget *new;
    GtkWidget *image;
    image = gtk_image_new_from_stock (image1, GTK_ICON_SIZE_LARGE_TOOLBAR);
    new = gtk_button_new();
    gtk_button_set_image (GTK_BUTTON (new), image);
    gtk_signal_connect(GTK_OBJECT(new), "clicked",
             GTK_SIGNAL_FUNC(callback), NULL);
    gtk_box_pack_start(GTK_BOX(container), new, TRUE, TRUE, 0);
    if (accel) {
        gtk_widget_add_accelerator(new, "clicked", accel_group,
                   accel, 0, GTK_ACCEL_VISIBLE);
    }
    gtk_widget_set_name(new, name);
    gtk_widget_show(new);
    return new;
}

static GtkWidget *add_menu_item(const gchar *label, MenuFunc callback,
                gint accel, GtkWidget *sub) {
    GtkWidget *itm;
    itm = gtk_menu_item_new_with_label(label);
    gtk_widget_set_name(itm, "mitem");
    gtk_menu_append(GTK_MENU(sub), itm);
    gtk_signal_connect_object(GTK_OBJECT(itm), "activate",
                GTK_SIGNAL_FUNC(callback), NULL);
    if (accel) {
        gtk_widget_add_accelerator(itm, "activate", accel_group,
                               accel, 0, GTK_ACCEL_VISIBLE);
    }
    gtk_widget_show(itm);
    return itm;
}

int main(int argc, char **argv) {
    struct sigaction act;
    gchar buf[STRBUFSIZE], wintitlestring[STRBUFSIZE];
    int c, j; /* for getopt */
    /* Set up gettext */
    setlocale(LC_ALL, "");
    bindtextdomain(PACKAGE, LOCALEDIR);
    textdomain(PACKAGE);
    /* Check that HOME is available */
    if (!getenv("HOME")) {
        fprintf(stderr, _("You must set $HOME to your home directory.\n"));
        return 20;
    }
    /* Get the CWD */
    if (getcwd(startup_cwd, STRBUFSIZE - 1)) {
        /* We left 1 space above for the /  */
        strcat(startup_cwd, "/");
    } else {
        /* Leave it blank; prepending it to paths won't do any damage then */
        *startup_cwd = '\0';
    }
    /* Set up the pipe used for the SIGCHLD handler */
    pipe(sigchld_pipe);
    /* Seed RNG */
    srand(time(0));
    /* Handle arguments */
    gtk_init(&argc, &argv);
    /* Read the gtkrc files */
    snprintf(buf, STRBUFSIZE, "%s/.xh/conf/gtkrc", getenv("HOME"));
    gtk_rc_parse(buf);
    /* Set our response to SIGCHLD */
    act.sa_handler = handle_sigchld;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NOCLDSTOP;
    sigaction(SIGCHLD, &act, NULL);
    /* Create the acceleration group */
    accel_group = gtk_accel_group_new();
    /* Create a window */
    snprintf(wintitlestring, STRBUFSIZE, "XH");
    /* Standard GTK initialisation */
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), wintitlestring);
    const gchar *wname = "xh";
    gtk_window_set_icon_name(GTK_WINDOW(window), wname);
    gtk_window_set_default_icon_name (wname);
    gtk_widget_set_name(window, "xhw");
    gtk_container_set_border_width (GTK_CONTAINER (window), 0);
    gtk_signal_connect(GTK_OBJECT(window), "destroy",
             GTK_SIGNAL_FUNC(handle_destroy), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "configure_event",
             GTK_SIGNAL_FUNC(handle_configure), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "window-state-event",
             GTK_SIGNAL_FUNC(handle_state), NULL);
    /* Listen to drag-and-drop */
    gtk_signal_connect(GTK_OBJECT(window), "drag_data_received",
             GTK_SIGNAL_FUNC(handle_dnd_drop), NULL);
    gtk_drag_dest_set(window, GTK_DEST_DEFAULT_MOTION |
             GTK_DEST_DEFAULT_HIGHLIGHT | GTK_DEST_DEFAULT_DROP,
             target_table, 1, GDK_ACTION_COPY | GDK_ACTION_MOVE);
    /* Visible UI */
    frameb = gtk_frame_new (NULL);
    gtk_container_add(GTK_CONTAINER(window), frameb);
    gtk_widget_set_name(frameb, "fbd");
    allbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(frameb), allbox);
    gtk_widget_set_name(allbox, "vbx");
    hnbox = gtk_handle_box_new();
    gtk_handle_box_set_handle_position(GTK_HANDLE_BOX(hnbox), GTK_POS_TOP);
    gtk_handle_box_set_snap_edge(GTK_HANDLE_BOX(hnbox), GTK_POS_TOP);
    gtk_container_set_border_width(GTK_CONTAINER (hnbox), 4);
    gtk_widget_set_name(hnbox, "hnb");
    gtk_box_pack_start(GTK_BOX(allbox), hnbox, FALSE, FALSE, 0);
    movebox = gtk_vbox_new(FALSE, 0);
    gtk_widget_set_name(movebox, "vubx");
    gtk_container_add(GTK_CONTAINER (hnbox), movebox);
    eventbox = gtk_event_box_new();
    gtk_widget_set_name(eventbox, "ebox");
    gtk_widget_add_events (eventbox, GDK_SCROLL_MASK);
    gtk_signal_connect(GTK_OBJECT(eventbox), "scroll-event",
             GTK_SIGNAL_FUNC(handle_ebox_scroll), NULL);
    gtk_signal_connect(GTK_OBJECT(eventbox), "button_press_event",
             GTK_SIGNAL_FUNC(handle_button_press), NULL);
    gtk_box_pack_start(GTK_BOX(movebox), eventbox, FALSE, FALSE, 0);
    statusbar = gtk_statusbar_new();
    gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), TRUE);
    gtk_widget_set_size_request(statusbar, 300, -1);
    gtk_container_add(GTK_CONTAINER(eventbox), statusbar);
    contextid = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "xh");
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), contextid, _("Welcome!"));
    gtk_widget_set_name(statusbar, "stbar");
    gtk_widget_show(statusbar);
    gtk_widget_show(eventbox);
    btbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(movebox), btbox, FALSE, FALSE, 0);
    gtk_widget_set_name(btbox, "btbox");
    btstop = add_button(GTK_STOCK_MEDIA_STOP, handle_stop, btbox, "xbstop", 
             GDK_s);
    btprev = add_button(GTK_STOCK_MEDIA_PREVIOUS, handle_prev, btbox, "xbprev",
             GDK_b);
    btplay = gtk_button_new();
    imbox = gtk_hbox_new (TRUE, 0);
    implay = gtk_image_new_from_stock (GTK_STOCK_MEDIA_PLAY,
                GTK_ICON_SIZE_LARGE_TOOLBAR);
    impause = gtk_image_new_from_stock (GTK_STOCK_MEDIA_PAUSE,
                GTK_ICON_SIZE_LARGE_TOOLBAR);
    gtk_box_pack_start (GTK_BOX (imbox), implay, TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (imbox), impause, TRUE, TRUE, 0);
    gtk_container_add (GTK_CONTAINER(btplay), imbox);
    gtk_signal_connect(GTK_OBJECT(btplay), "clicked",
                     GTK_SIGNAL_FUNC(handle_pause), NULL);
    gtk_box_pack_start(GTK_BOX(btbox), btplay, TRUE, TRUE, 0);
    gtk_widget_add_accelerator(btplay, "clicked", accel_group,
                             GDK_p, 0, GTK_ACCEL_VISIBLE);
    gtk_widget_set_name(btplay, "xbplay");
    gtk_widget_show(implay);
    gtk_widget_show(impause);
    gtk_widget_show(imbox);
    gtk_widget_show(btplay);
    gtk_widget_set_can_focus (btplay , TRUE);
    gtk_widget_hide(impause);
    btnext = add_button(GTK_STOCK_MEDIA_NEXT, handle_next, btbox, "xbnext", 
             GDK_n);     
    btlist = gtk_button_new();
    imbox2 = gtk_hbox_new (TRUE, 0);
    imlists = gtk_image_new_from_stock (GTK_STOCK_JUSTIFY_FILL,
                GTK_ICON_SIZE_LARGE_TOOLBAR);
    imlisth = gtk_image_new_from_stock (GTK_STOCK_JUSTIFY_LEFT,
                GTK_ICON_SIZE_LARGE_TOOLBAR);
    gtk_box_pack_start (GTK_BOX (imbox2), imlists, TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (imbox2), imlisth, TRUE, TRUE, 0);
    gtk_container_add (GTK_CONTAINER(btlist), imbox2);
    gtk_signal_connect(GTK_OBJECT(btlist), "clicked",
                     GTK_SIGNAL_FUNC(handle_minitoggle), NULL);
    gtk_box_pack_start(GTK_BOX(btbox), btlist, TRUE, TRUE, 0);
    gtk_widget_add_accelerator(btlist, "clicked", accel_group,
                             GDK_i, 0, GTK_ACCEL_VISIBLE);
    gtk_widget_set_name(btlist, "xblist");
    gtk_widget_show(imbox2);
    gtk_widget_show(btlist);
    gtk_widget_show(btbox);
    gtk_widget_show(movebox);
    gtk_widget_show(hnbox);
    listbox = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(listbox),
                                  GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(listbox),
                                       GTK_SHADOW_NONE);
    gtk_box_pack_start(GTK_BOX(allbox), listbox, TRUE, TRUE, 0);
    gtk_widget_set_name(listbox, "listbox");
    /* Create the popup menu */
    popupmenu = gtk_menu_new();
    g_object_set(G_OBJECT(popupmenu), "reserve-toggle-size", FALSE, NULL);
    gtk_widget_set_name(popupmenu, "xhmenu");
    itemdel = add_menu_item(_("Remove song"), handle_menu_delete, 0, popupmenu);
    add_menu_item(_("Add song"), handle_menu_add, GDK_a, popupmenu);
#ifdef HAVE_NFTW
    add_menu_item(_("Add directory"), handle_menu_add_dir, GDK_d, popupmenu);
#endif
    msep1 = gtk_separator_menu_item_new();
    gtk_widget_set_name(msep1, "msep");
    gtk_menu_append(GTK_MENU(popupmenu), msep1);
    gtk_widget_show(msep1);
    /* Playlist submenu */
    subl = gtk_menu_new();
    g_object_set(G_OBJECT(subl), "reserve-toggle-size", FALSE, NULL);
    gtk_widget_set_name(subl, "menusub");
    add_menu_item(_("Load"), handle_menu_load, GDK_l, subl);
    add_menu_item(_("Save"), handle_menu_save, GDK_v, subl);
    add_menu_item(_("Clear"), handle_menu_clear, GDK_c, subl);
    iteml = gtk_menu_item_new_with_label(_("Playlist"));
    gtk_widget_set_name(iteml, "mitem");
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (iteml), subl);
    gtk_menu_shell_append (GTK_MENU_SHELL (popupmenu), iteml);
    gtk_widget_show(iteml);
    /* Sort submenu */
    subr = gtk_menu_new();
    g_object_set(G_OBJECT(subr), "reserve-toggle-size", TRUE, NULL);
    gtk_widget_set_name(subr, "menusub");
    /* radiomenubuttons */
    GSList *group = NULL;
    rb0 = gtk_radio_menu_item_new_with_label (group, (_("normal")));
    group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (rb0));
    gtk_signal_connect_object(GTK_OBJECT(rb0), "activate",
                GTK_SIGNAL_FUNC(handle_menu_sort_normal), NULL);
    gtk_widget_add_accelerator(rb0, "activate", accel_group, GDK_z, 0, 
                               GTK_ACCEL_VISIBLE);
    gtk_menu_append(GTK_MENU(subr), rb0);
    rb1 = gtk_radio_menu_item_new_with_label (group, (_("name")));
    group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (rb1));
    gtk_signal_connect_object(GTK_OBJECT(rb1), "activate",
                GTK_SIGNAL_FUNC(handle_menu_sort_name), NULL);
    gtk_widget_add_accelerator(rb1, "activate", accel_group, GDK_o, 0, 
                               GTK_ACCEL_VISIBLE);
    gtk_menu_append(GTK_MENU(subr), rb1);
    rb2 = gtk_radio_menu_item_new_with_label (group, (_("swapped name")));
    group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (rb2));
    gtk_signal_connect_object(GTK_OBJECT(rb2), "activate",
                GTK_SIGNAL_FUNC(handle_menu_sort_swapped_name), NULL);
    gtk_widget_add_accelerator(rb2, "activate", accel_group, GDK_w, 0, 
                               GTK_ACCEL_VISIBLE);
    gtk_menu_append(GTK_MENU(subr), rb2);
    rb3 = gtk_radio_menu_item_new_with_label (group, (_("mtime")));
    group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (rb3));
    gtk_signal_connect_object(GTK_OBJECT(rb3), "activate",
                GTK_SIGNAL_FUNC(handle_menu_sort_mtime), NULL);
    gtk_widget_add_accelerator(rb3, "activate", accel_group, GDK_t, 0, 
                               GTK_ACCEL_VISIBLE);
    gtk_menu_append(GTK_MENU(subr), rb3);
    gtk_widget_set_name(rb0, "rmb");
    gtk_widget_set_name(rb1, "rmb");
    gtk_widget_set_name(rb2, "rmb");
    gtk_widget_set_name(rb3, "rmb");
    gtk_widget_show(rb0);
    gtk_widget_show(rb1);
    gtk_widget_show(rb2);
    gtk_widget_show(rb3);
    msep3 = gtk_separator_menu_item_new();
    gtk_widget_set_name(msep3, "msepsub");
    gtk_menu_append(GTK_MENU(subr), msep3);
    gtk_widget_show(msep3);
    ckm = gtk_check_menu_item_new_with_label (_("random"));
    gtk_widget_set_name(ckm, "cmb");
    gtk_signal_connect_object(GTK_OBJECT(ckm), "activate",
                GTK_SIGNAL_FUNC(handle_randomtoggle), NULL);
    gtk_widget_add_accelerator(ckm, "activate", accel_group, GDK_r, 0,
                                GTK_ACCEL_VISIBLE);
    gtk_menu_append(GTK_MENU(subr), ckm);
    gtk_widget_show(ckm);
    ckl = gtk_check_menu_item_new_with_label (_("loop"));
    gtk_widget_set_name(ckl, "cmb");
    gtk_signal_connect_object(GTK_OBJECT(ckl), "activate",
                GTK_SIGNAL_FUNC(handle_looptoggle), NULL);
    gtk_widget_add_accelerator(ckl, "activate", accel_group, GDK_u, 0,
                                GTK_ACCEL_VISIBLE);
    gtk_menu_append(GTK_MENU(subr), ckl);
    gtk_widget_show(ckl);
    itemr = gtk_menu_item_new_with_label(_("Sort"));
    gtk_widget_set_name(itemr, "mitem");
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (itemr), subr);
    gtk_menu_shell_append (GTK_MENU_SHELL (popupmenu), itemr);
    gtk_widget_show(itemr);
    msep2 = gtk_separator_menu_item_new();
    gtk_widget_set_name(msep2, "msep");
    gtk_menu_append(GTK_MENU(popupmenu), msep2);
    gtk_widget_show(msep2);
    add_menu_item(_("Setup"), handle_menu_prefs, 0, popupmenu);
    add_menu_item(_("Quit"), handle_menu_quit, GDK_q, popupmenu);
    /* Initialise the config system */
    init_config_commands();
    /* Read the config file for the first time */
    read_config_file(0);
    /* Parse arguments */
    while (1) {
        char *argstring = "ahqcfdxrg";
#ifndef NOGETOPTLONG
        static struct option long_options[] = {
            {"autostart",0,0,0}, /* -a */
            {"help",0,0,0}, /* -h */
            {"hideplayeroutput",0,0,0}, /* -q */
            {"nocheckfiles",0,0,0}, /* -c */
            {"cmdsongs",0,0,0}, /* -f */
            {"cmddirs",0,0,0}, /* -d */
            {"removeplayed",0,0,0}, /* -x */
            {"roam",0,0,0}, /* -r */
            {"cmdguess",0,0,0}, /* -g */
            {0,0,0,0}
        };
        j = 0;
        c = getopt_long(argc, argv, argstring, long_options, &j);
#else
        c = getopt(argc, argv, argstring);
#endif
        if (c == -1)
            break; /* out of the while; done with the args */
#ifndef NOGETOPTLONG
        /* A long option; look it up in the arguments string */
        if (c == 0) {
            char *p = argstring;
            while (j--) {
                p++;
                if (*p == ':')
                    p++;
            }
            c = *p;
        }
#endif
        switch(c) {
            case 'a': /* -a: automatic play on start */
                mode_play_on_start = 1;
                break;
            case 'q': /* -q: hide player output */
                mode_hide_player_output = 1;
                break;
            case 'c': /* -c: don't bother stat()ing files on startup */
                mode_no_check_files = 1;
                break;
            case 'f': /* -f: songs (rather than playlist) on command line */
                mode_commandline_songs = 1;
                break;
            case 'd': /* -d: directories (rather than playlists) on command line */
                mode_commandline_dirs = 1;
                break;
            case 'x': /* -x: delete when played */
                mode_delete_when_played = 1;
                break;
            case 'r': /* -r: do not persist playlist */
                mode_persist_playlist = 0;
                break;
            case 'g': /* -g: guess command line arguments */
                mode_commandline_guess = 1;
                break;
            case 'h':
                g_print("\n XH ver.03, GNU GPLv3 \n");
                return 0;
            default:
                return 20;
        }
    }
    {
        GtkCellRenderer *renderer;
        GtkTreeSelection *select;
        filestore = gtk_list_store_new(NUM_COLUMNS, G_TYPE_POINTER,
                   G_TYPE_STRING);
        filelist = gtk_tree_view_new_with_model(GTK_TREE_MODEL(filestore));
        g_signal_connect(G_OBJECT(filestore), "row-deleted",
                G_CALLBACK(handle_row_deleted), NULL);
        g_signal_connect(G_OBJECT(filestore), "row-inserted",
                G_CALLBACK(handle_row_inserted), NULL);
        g_object_unref(G_OBJECT(filestore));
        gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(filelist), FALSE);
        renderer = gtk_cell_renderer_text_new();
        column = gtk_tree_view_column_new_with_attributes("XH", renderer,
                              "text", TITLE_COLUMN, NULL);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
        gtk_tree_view_set_reorderable(GTK_TREE_VIEW(filelist), TRUE);
        gtk_tree_view_append_column(GTK_TREE_VIEW(filelist), column);
        select = gtk_tree_view_get_selection(GTK_TREE_VIEW(filelist));
        gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
        g_signal_connect(G_OBJECT(filelist), "button_press_event",
                G_CALLBACK(handle_button_press), NULL);
        g_signal_connect(G_OBJECT(filelist), "row-activated",
                G_CALLBACK(handle_row_activated), NULL);
    }
    /* Since we add lots of items to the file list, we don't want it
       requesting a really really tall size for itself */
    gtk_widget_set_size_request(GTK_WIDGET(filelist), -1, 0);
    gtk_widget_set_name(filelist, "filelist");
    gtk_container_add (GTK_CONTAINER(listbox), filelist);
    /* Load the playlist if no other args are given */
    snprintf(buf, STRBUFSIZE, "%s/.xh/playlists/playlist", getenv("HOME"));
    if (mode_persist_playlist && optind >= argc)
        read_playlist(buf, 1);
    /* Deal with non-options */
    while (optind < argc) {
        char *name = argv[optind++];
        if (mode_commandline_songs) {
            add_file(name);
        } else if (mode_commandline_dirs) {
            add_directory(name);
        } else {
            if (mode_commandline_guess) {
                fileaction *fa = get_action(name);
                if (fa) {
                    add_file(name);
                } else if (filetype(name) == 2) {
                    add_directory(name);
                } else {
                    read_playlist(name, 0);
                }
            } else {
                read_playlist(name, 0);
            }
        }
    }
    gtk_widget_show(filelist);
    gtk_widget_show(listbox);
    gtk_widget_show(allbox);
    /* Attach the acceleration group to the window */
    gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
    /* Read the config file for the second time */
    read_config_file(1);
    /* Set the status */
    snprintf(buf, STRBUFSIZE, _("%d files in list."), listcount);
    status(buf);
    /* Set the window attributes */
    read_window_state();
    if (mode_window_xy) {
        gtk_window_move(GTK_WINDOW(window), attribute_xpos, attribute_ypos);
    } else {
        gtk_window_set_position (GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
    }
    if (mode_window_wh) {
        if (attribute_width && attribute_height)
            gtk_window_resize(GTK_WINDOW(window), attribute_width,
                  attribute_height);
    }
    gtk_window_set_decorated(GTK_WINDOW (window), mode_decorated);
    set_mini();
    read_window_prop();
    /* Update optionmenu */
     if (mode_sort_type == 0) {
         gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (rb0), TRUE);
     } else if (mode_sort_type == 1) {
         gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (rb1), TRUE);
     } else if (mode_sort_type == 2) {
         gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (rb2), TRUE);
     } else if (mode_sort_type == 3) {
         gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (rb3), TRUE);
     }
    /* Update checkboxes */
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(ckm), mode_random);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(ckl), mode_loop);
    gtk_widget_show(frameb);
    gtk_widget_show(window);
    /* Set initial keyboard focus */
    gtk_widget_grab_focus(btplay);
    /* If mode_play_on_start, start playing */
    if (mode_play_on_start)
        pick_next();
    /* Make GDK monitor the pipe */
    gdk_input_add(sigchld_pipe[0], GDK_INPUT_READ, sigchld_backend, NULL);
    /* Main loop */
    gtk_main();
    /* Note that the interface doesn't exist any more, so you can't do
       anything interesting here */
    return 0;
}

